from smah.database import Database, Migration
import argparse


def parse_arguments():
    parser = argparse.ArgumentParser(description="Migration tool for managing database migrations.")

    # Create subparsers for each command (migrate, rollback, status)
    subparsers = parser.add_subparsers(dest="command", help="Migration commands")

    # Migrate command
    migrate_parser = subparsers.add_parser("migrate", help="Apply migrations")
    migrate_parser.add_argument("--to", type=str, help="Migrate up to a specific migration name")
    migrate_parser.add_argument("--count", type=int, help="Apply a specified number of migrations")
    migrate_parser.add_argument("--reset-checksums",
                                action=argparse.BooleanOptionalAction,
                                help="Reset migration checksums on mismatch",
                                default=False)

    # Rollback command
    rollback_parser = subparsers.add_parser("rollback", help="Rollback migrations")
    rollback_parser.add_argument("--to", type=str, help="Rollback to a specific migration name")
    rollback_parser.add_argument("--count", type=int, help="Rollback a specified number of migrations")
    rollback_parser.add_argument("--reset-checksums",
                                action=argparse.BooleanOptionalAction,
                                help="Reset migration checksums on mismatch",
                                default=False)


    # Status command
    subparsers.add_parser("status", help="Show the current migration status")

    # Status command
    create_migration_parser = subparsers.add_parser("create", help="Show the current migration status")
    create_migration_parser.add_argument(dest="name", type=str, help="Name of the migration")

    # database argument
    parser.add_argument("--database", type=str, help="Path to the database file")

    return parser.parse_args()

def main():
    args = parse_arguments()
    database = Database(args)

    if args.command == "migrate":
        Migration.migrate(database, args)
    elif args.command == "rollback":
        Migration.rollback(database, args)
    elif args.command == "status":
        Migration.status(database)
    elif args.command == "create":
        Migration.create(args.name)



if __name__ == "__main__":
    main()


import argparse
import hashlib
import os
import time
import importlib
import textwrap
import sqlite3
from smah.database.database import Database

class Migration:
    # Migrations are stored in the `migrations` directory
    MIGRATIONS_DIR = os.path.join(os.path.dirname(__file__), "migrations")

    def __init__(self):
        pass


    @staticmethod
    def get_schema_migrations(database: Database):
        cursor = database.connection.cursor()
        create_table = textwrap.dedent(
            """
            CREATE TABLE IF NOT EXISTS schema_migrations(
                migration VARCHAR(255) PRIMARY KEY, 
                checksum CHAR(32), 
                applied BOOLEAN DEFAULT FALSE, 
                created_at TIMESTAMP DEFAULT CURRENT_TIMESTAMP, 
                modified_at TIMESTAMP DEFAULT CURRENT_TIMESTAMP
            )
            """
        ).strip()
        cursor.execute(create_table)

        cursor.execute("SELECT migration, checksum, applied, created_at, modified_at FROM schema_migrations ORDER BY migration ASC")
        result = cursor.fetchall()
        cursor.close()
        migrations = []
        for row in result:
            migration, checksum, applied, created_at, modified_at = row
            migrations.append(
                {
                    'migration': migration,
                    'checksum': checksum,
                    'applied': applied == 1,
                    'created_at': created_at,
                    'modified_at': modified_at
                }
            )
        return migrations

    @staticmethod
    def get_migrations():
        migrations = []
        os.makedirs(Migration.MIGRATIONS_DIR, exist_ok=True)
        for migration in os.listdir(Migration.MIGRATIONS_DIR):
            if migration.endswith(".py"):
                digest = hashlib.md5(open(os.path.join(Migration.MIGRATIONS_DIR, migration), "rb").read()).hexdigest()
                migrations.append({'file': migration, 'checksum': digest})
        return migrations

    @staticmethod
    def apply_migration(database: Database, migration: dict):
        cursor = database.connection.cursor()
        module = importlib.import_module(f"smah.database.migrations.{migration['file'][:-3]}")
        cursor.execute("BEGIN TRANSACTION")
        try:
            module.up(cursor)
            cursor.execute(
            """
                INSERT INTO schema_migrations (migration, checksum, applied) 
                VALUES (?, ?, ?)
                ON CONFLICT(migration) DO UPDATE SET
                    checksum = excluded.checksum,
                    applied = excluded.applied
                """,
                (migration['file'], migration['checksum'], True)
            )
            cursor.execute("COMMIT")
            cursor.close()
            print(f"Applied Migration {migration['file']}")
        except Exception as e:
            cursor.execute("ROLLBACK")
            cursor.close()
            print(f"Error Applying Migration {migration['file']}: {e}")
            exit(2)

    @staticmethod
    def rollback_migration(database: Database, migration: dict):
        cursor = database.connection.cursor()
        module = importlib.import_module(f"smah.database.migrations.{migration['file'][:-3]}")
        cursor.execute("BEGIN TRANSACTION")
        try:
            module.down(cursor)
            cursor.execute(
                """
                INSERT INTO schema_migrations (migration, checksum, applied) 
                VALUES (?, ?, ?)
                ON CONFLICT(migration) DO UPDATE SET
                    checksum = excluded.checksum,
                    applied = excluded.applied
                """,
                (migration['file'], migration['checksum'], False)
            )
            cursor.execute("COMMIT")
            cursor.close()
            print(f"Reverted Migration '{migration['file']}'")
        except Exception as e:
            cursor.execute("ROLLBACK")
            cursor.close()
            print(f"Error Reverting Migration '{migration['file']}': {e}")
            exit(2)

    @staticmethod
    def migrate(database: Database, args: argparse.Namespace):
        count = 0
        available_migrations = Migration.get_migrations()
        tracked_migrations = {m['migration']: m for m in Migration.get_schema_migrations(database)}

        for migration in available_migrations:
            # Skip migrations that have already been applied
            tracked = tracked_migrations.get(migration['file'])
            if tracked and tracked['applied']:
                if tracked['checksum'] != migration['checksum']:
                    print(f"Checksum Mismatch in {migration['file']}: expected {tracked['checksum']} but got {migration['checksum']}")
                    if args.reset_checksums:
                        cursor = database.connection.cursor()
                        cursor.execute("UPDATE schema_migrations SET checksum = ? WHERE migration = ?", (migration['checksum'], migration['file']))
                        cursor.close()
                    else:
                        exit(1)
                if args.to and args.to == migration['file']:
                    print(f"Migration Complete: changes applied {count}")
                    exit(0)
            else:
                Migration.apply_migration(database, migration)
                count += 1
                if args.count and count >= args.count:
                    print(f"Migration Complete: changes applied {count}")
                    exit(0)
                if args.to and args.to == migration['file']:
                    print(f"Migration Complete: changes applied {count}")
                    exit(0)
        if count == 0:
            print("No Migrations Pending")
        else:
            print(f"Migration Complete: changes applied {count}")

    @staticmethod
    def rollback(database: Database, args: argparse.Namespace):
        count = 0
        available_migrations = {m['file']: m for m in Migration.get_migrations()}
        tracked_migrations = [m for m in Migration.get_schema_migrations(database) if m['applied'] and m['migration'] in available_migrations]
        tracked_migrations.reverse()
        if args.to:
            # Verify to target is applied
            available = False
            for migration in tracked_migrations:
                if migration['migration'] == args.to:
                    available = True
                    break
            if not available:
                print(f"Migration --to '{args.to}' not applied or available")
                exit(1)
        for migration in tracked_migrations:
            Migration.rollback_migration(database, available_migrations[migration['migration']])
            count += 1
            if args.count and count >= args.count:
                print(f"Rollback Complete: changes applied {count}")
                exit(0)
            if args.to and args.to == migration['migration']:
                print(f"Rollback Complete: changes applied {count}")
                exit(0)
        if count == 0:
            print("No Migrations Available for Rollback")
        else:
            print(f"Rollback Complete #{count} migrations reverted.")

    @staticmethod
    def status(database: Database):
        available_migrations = Migration.get_migrations()
        tracked_migrations = {m['migration']: m for m in Migration.get_schema_migrations(database)}
        if not available_migrations:
            print("No Migrations Available")
            exit(0)
        out = ""
        for migration in available_migrations:
            migration['applied'] = False
            migration['checksum_mismatch'] = False
            tracked = tracked_migrations.get(migration['file'])
            if tracked:
                migration['applied'] = tracked['applied']
                if tracked['checksum'] != migration['checksum']:
                    migration['checksum_mismatch'] = True
            out += f"{migration['file']} {'(applied)' if migration['applied'] else ''} {'(checksum mismatch)' if migration['checksum_mismatch'] else ''}\n"
        print(f"Migrations:\n{out}")

    @staticmethod
    def create(name: str) -> None:
        epoch = int(time.time())
        template = textwrap.dedent(
            """
            def up(cursor):
                \"\"\"
                Apply schema.
                \"\"\"
                pass
                
            def down(cursor):
                \"\"\"
                Rollback schema.
                \"\"\"
                pass    
         
            """).strip()
        mf = os.path.join(Migration.MIGRATIONS_DIR, f"{epoch}_{name}.py")
        os.makedirs(Migration.MIGRATIONS_DIR, exist_ok=True)
        with open(mf, "w") as file:
            file.write(template)
        print(f"Created migration: {mf}")




import textwrap
from typing import Optional, Tuple

import yaml


class BaseInfo:
    CONFIG_VSN = "0.0.1"

    @staticmethod
    def info()  -> Optional[Tuple]:
        return None

    @staticmethod
    def config_vsn():
        return BaseInfo.CONFIG_VSN


    def __init__(self, kind: str, config_data = None, fetch=False):
        config_data = config_data or {}
        self.kind = kind
        self.source = None
        self.details = None
        self.vsn = None

        if fetch:
            self.vsn = self.config_vsn()
            details = self.info()
            if details:
                self.source, self.details = details
            else:
                self.source = "Unsupported"
        elif config_data is not None:
            self.vsn = config_data.get("vsn", self.config_vsn())
            self.source = config_data.get("source")
            self.details = config_data.get("details")

    def is_configured(self):
        """
        Checks if the OS details are configured.

        Returns:
            bool: True if the details are configured, False otherwise.
        """
        if not self.kind:
            return False
        return True

    def to_yaml(self, options = None):
        """
        Converts the OS details to a YAML-compatible dictionary.

        Returns:
            dict: The OS details in YAML format.
        """

        return {
            "vsn": self.config_vsn(),
            "kind": self.kind,
            "source": self.source,
            "details": self.details,
        }

    def show(self, options = None):
        if self.details:
            details = textwrap.indent("\n".join([f"- {k}: {v}" for k,v in self.details.items()]), "  ")
        else:
            details = None

        template = textwrap.dedent(
            """
            - kind: {kind}
            - source: {source}
            - details:            
            {details}
            """
        ).strip().format(
            kind=self.kind,
            source=self.source,
            details=details
        )
        return template




from .base_stats import BaseStats
import textwrap
import psutil
import datetime

class CpuStats(BaseStats):
    """
    Represents CPU statistics.

    Attributes:
        time_stamp (datetime): The timestamp of the last update.
        cpu_count (int): The last recorded CPU count.
        cpu_freq (float): The last recorded CPU frequency.
        cpu_percent (float): The last recorded CPU usage percentage.
    """

    @staticmethod
    def cpu_info(reading):
        """
        Retrieves CPU information based on the specified reading type.

        Args:
            reading (str): The type of CPU information to retrieve.

        Returns:
            int or float: The requested CPU information.
        """
        try:
            if reading == "count":
                return psutil.cpu_count(logical=True)
            elif reading == "freq.current":
                return round(psutil.cpu_freq().current, 2)
            elif reading == "percent":
                return round(psutil.cpu_percent(), 2)
        except:
            return None

    def __init__(self):
        """
        Initializes the CpuStats instance.
        """
        super().__init__()
        self.cpu_count = None
        self.cpu_freq = None
        self.cpu_percent = None

    def update(self):
        """
        Updates the CPU statistics.
        """
        self.time_stamp = datetime.datetime.now()
        self.cpu_count = self.cpu_info("count")
        self.cpu_freq = self.cpu_info("freq.current")
        self.cpu_percent = self.cpu_info("percent")

    def readings(self, threshold=100):
        """
        Retrieves the current CPU readings, updating if necessary.

        Args:
            threshold (int): The threshold in microseconds.

        Returns:
            dict: The current CPU readings.
        """
        if self.stale(threshold):
            self.update()
        return {
            "time": self.time_stamp,
            "cpu_count": self.cpu_count,
            "cpu_freq": self.cpu_freq,
            "cpu_percent": self.cpu_percent
        }

    def show(self, options = None):
        if self.stale():
            self.update()
        template = textwrap.dedent(
            """
            - time: {time}
            - count: {count}
            - freq: {freq}
            - percent: {percent}
            """
        ).format(
            time=self.time_stamp,
            count=self.cpu_count,
            freq=self.cpu_freq,
            percent=self.cpu_percent
        )



from typing import Optional

import rich.box
from rich.markdown import Markdown
from rich.panel import Panel
from rich.prompt import Confirm, Prompt
from .user import User
from smah.console import std_console, err_console, prompt_choice, prompt_string
import textwrap

def user_terminal_configurator(user: Optional[User]) -> User:
    std_console.print(Markdown("## Configure User"))
    user = user or User()

    if user.is_configured():
        show(user)
        if not Confirm.ask("edit?", default=False):
            return user
        else:
            std_console.print("\n")

    while(True):
        user = prompt(user)

        std_console.print("\n")
        show(user, label = "Confirm User Settings", border_style="green bold")
        if Confirm.ask("confirm?", default=True):
            return user

def prompt(user: User) -> User:
    user.name = prompt_string("Name", user.name, label="What is your name?")
    user.system_admin_experience = prompt_choice(
        "Experience Level",
        user.system_admin_experience,
        label = "What is your experience level?",
        options = ["Novice", "Beginner", "Intermediate", "Advanced", "Expert"]
    )
    user.role = prompt_choice(
        "Role",
        user.role,
        label = "What is your role?",
        options = ["Developer", "Administrator", "Student", "User"],
        other=True
    )
    user.about = prompt_string(
        "About",
        user.about,
        label="Tell us about yourself"
    )
    return user

def show(user: User, border_style="white bold", label = "User Settings"):
    body = user.show()
    template = Markdown(body)
    template = Panel(template, title=label, border_style=border_style, box=rich.box.ROUNDED)
    std_console.print(template)


import textwrap
from enum import Enum
from typing import Optional
from lxml import etree

html_tags = [
    "a",
    "abbr",
    "acronym",
    "address",
    "applet",
    "area",
    "article",
    "aside",
    "audio",
    "b",
    "base",
    "basefont",
    "bdo",
    "big",
    "blockquote",
    "body",
    "br",
    "button",
    "canvas",
    "caption",
    "center",
    "cite",
    "code",
    "col",
    "colgroup",
    "datalist",
    "dd",
    "del",
    "dfn",
    "div",
    "dl",
    "dt",
    "em",
    "embed",
    "fieldset",
    "figcaption",
    "figure",
    "font",
    "footer",
    "form",
    "frame",
    "frameset",
    "head",
    "header",
    "h1",
    "h2",
    "h3",
    "h4",
    "h5",
    "h6",
    "hr",
    "html",
    "i",
    "iframe",
    "img",
    "input",
    "ins",
    "kbd",
    "label",
    "legend",
    "li",
    "link",
    "main",
    "map",
    "mark",
    "meta",
    "meter",
    "nav",
    "noscript",
    "object",
    "ol",
    "optgroup",
    "option",
    "p",
    "param",
    "pre",
    "progress",
    "q",
    "s",
    "samp",
    "script",
    "section",
    "select",
    "small",
    "source",
    "span",
    "strike",
    "strong",
    "style",
    "sub",
    "sup",
    "table",
    "tbody",
    "td",
    "textarea",
    "tfoot",
    "th",
    "thead",
    "time",
    "title",
    "tr",
    "u",
    "ul",
    "var",
    "video",
    "wbr",
]

class ThoughtType(Enum):
    OTHER = 0
    THINKING = 1
    QUESTION = 2
    ASSUMPTION = 3
    INNER_CRITIC = 4
    TANGENT = 5

class TagBase(etree.ElementBase):
    def extract_child(self, tag: str):
        child = self.cssselect(tag)
        if len(child) > 0:
            child = child[0]
            return ResponseParser.unescape_response(child.text.strip())
        return None

class SetConditionTag(TagBase):
    @property
    def name(self):
        return ResponseParser.unescape_response(self.get("name"))

    @property
    def prompt(self):
        prompt = self.extract_child("prompt")
        if prompt:
            return prompt
        else:
            name = ResponseParser.unescape_response(self.get("name"))
            return f"Set {name}"

    @property
    def choices(self):
        choices = []
        x = self.cssselect("choices choice")
        if len(x) > 0:
            for c in x:
                c_value = ResponseParser.unescape_response(c.get("value"))
                c_label = ResponseParser.unescape_response(c.text or c_value)
                i = None
                c_user = c.get("data-user") == "true"
                if c_user:
                    i = {
                        'prompt': ResponseParser.unescape_response(c.get("data-prompt") or f"Enter {c_label}"),
                        'required': c.get("data-required") == "true",
                        'check': ResponseParser.unescape_response(c.get("data-check"))
                    }
                choices.append({
                    'value': c_value,
                    'label': c_label,
                    'prompt': i
                })
        return choices








class ThoughtTag(TagBase):
    THOUGHT_TYPES = {
        "tangent": ThoughtType.TANGENT,
        "thinking": ThoughtType.THINKING,
        "question": ThoughtType.QUESTION,
        "assumption": ThoughtType.ASSUMPTION,
        "inner-critic": ThoughtType.INNER_CRITIC
    }


    def _init(self):
        self.tag = "cot"

    @property
    def type(self):
        return self.THOUGHT_TYPES.get(self.get("type"), ThoughtType.OTHER)

    @property
    def thought(self):
        return ResponseParser.unescape_response(self.text)

    @property
    def markdown(self):
        type = self.THOUGHT_TYPES.get(self.get("type"), ThoughtType.OTHER)
        if type == ThoughtType.THINKING:
            type = "Thinking"
        elif type == ThoughtType.QUESTION:
            type = "Question"
        elif type == ThoughtType.ASSUMPTION:
            type = "Assumption"
        elif type == ThoughtType.INNER_CRITIC:
            type = "Inner-Critic"
        elif type == ThoughtType.TANGENT:
            type = "Tangent"
        else:
            type = "Other"
        body = self.text or ""
        body = ResponseParser.unescape_response(body.replace("`","'"))
        return f"`{type}: {body}`"

class ExecTag(TagBase):
    def _init(self):
        self.tag = "exec"

    @property
    def shell(self):
        return self.get("shell")

    @property
    def exec_if(self):
        return ResponseParser.unescape_response(self.get("exec-if"))

    @property
    def title(self):
        return self.extract_child("title") or "Run Command"

    @property
    def purpose(self):
        return self.extract_child("purpose") or "No Purpose Provided"

    @property
    def command(self):
        return self.extract_child("command")

    @property
    def markdown(self):
        title = self.extract_child("title") or "Run Command"
        title = textwrap.indent(title, "# ")
        shell = self.get("shell")
        command = self.extract_child("command") or ""

        template = textwrap.dedent(
            """
            ```{shell}
            {title}
            
            {command}
            ```
            """
        ).strip().format(
            title=title,
            shell=shell,
            command=command
        )
        return template


class SmahLookup(etree.CustomElementClassLookup):
    def lookup(self, node_type, document, namespace, name):
        if node_type == "element":
            if name == "exec":
                return ExecTag
            elif name == "cot":
                return ThoughtTag
            elif name == "set-condition":
                return SetConditionTag
        return None

class ResponseParser:
    def __init__(self):
        pass

    @staticmethod
    def replace_tag(elem: any, replace: Optional[str] = None, tail: Optional[str] = None):
        ps = elem.getprevious()
        p = elem.getparent()
        if ps is not None:
            t = (ps.tail or "")
            l = t.split("\n")
            if len(l) > 0:
                r = len(l[-1])
                if replace:
                    t += textwrap.indent(replace, " " * r).lstrip()
                t += (tail or "")
            else:
                t += ((replace or "").lstrip() + (tail or ""))
            ps.tail = t

        elif p is not None:
            t = (p.text or "")
            l = t.split("\n")
            if len(l) > 0:
                r = len(l[-1])
                if replace:
                    t += textwrap.indent(replace, " " * r).lstrip()
                t += (tail or "")
            else:
                t += ((replace or "").lstrip() + (tail or ""))
            p.text = t
        if p is not None:
            p.remove(elem)


    @staticmethod
    def extract_conditions(response: str, options: Optional[dict] = None) -> Optional[list]:
        options = options or {}
        parser = etree.XMLPullParser(recover=True)
        parser.set_element_class_lookup(SmahLookup())
        events = parser.read_events()
        response = ResponseParser.escape_response(response)
        parser.feed("<smah-msg>" + response + "</smah-msg>")
        response = []
        for action, elem in events:
            if action == "end":
                if isinstance(elem, SetConditionTag):
                    c = {
                        'name': ResponseParser.unescape_response(elem.name),
                        'prompt': ResponseParser.unescape_response(elem.prompt),
                        'choices': elem.choices
                    }
                    response.append(c)
        return response

    @staticmethod
    def extract_commands(response: str, options: Optional[dict] = None) -> Optional[list]:
        options = options or {}
        parser = etree.XMLPullParser(recover=True)
        parser.set_element_class_lookup(SmahLookup())
        events = parser.read_events()

        response = ResponseParser.escape_response(response)
        parser.feed("<smah-msg>" + response + "</smah-msg>")
        commands = []
        for action, elem in events:
            if action == "end":
                if isinstance(elem, ExecTag):
                    conditions = options.get("conditions") or {}
                    # include operator and system
                    c = elem.exec_if
                    c = ResponseParser.unescape_response(c)
                    if c is None or eval(c, conditions):
                        c = {
                            'title': ResponseParser.unescape_response(elem.title),
                            'purpose': ResponseParser.unescape_response(elem.purpose),
                            'command': ResponseParser.unescape_response(elem.command),
                            'shell': ResponseParser.unescape_response(elem.shell)
                        }
                        commands.append(c)
                    else:
                        print(f"Skipping command: {elem.title} due to falsy condition: {c}")
        return commands

    @staticmethod
    def escape_response(response: str) -> str:
        response = response.replace("<", ":_smah_lt_:")
        response = response.replace("&amp;", ":_smah_amp_amp_:")
        response = response.replace("&", ":_smah_amp_:")

        for tag in ["smah-", "cot", "exec", "prompt", "title", "command", "set-condition", "choices", "choice"]:
            response = response.replace(f":_smah_lt_:{tag}", f"<{tag}")
            response = response.replace(f":_smah_lt_:/{tag}", f"</{tag}")

        for tag in html_tags:
            response = response.replace(f":_smah_lt_:{tag}", f"<{tag}")
            response = response.replace(f":_smah_lt_:/{tag}", f"</{tag}")


        return response

    @staticmethod
    def unescape_response(response: Optional[str]) -> Optional[str]:
        if response is None:
            return None
        response = response.replace(":_smah_lt_:", "<")
        response = response.replace(":_smah_amp_amp_:", "&amp;")
        response = response.replace(":_smah_amp_:", "&")

        return response

    @staticmethod
    def to_markdown(response: str, options: Optional[dict] = None) -> str:
        options = options or {}
        parser = etree.XMLPullParser(recover=True, encoding="utf-8")
        parser.set_element_class_lookup(SmahLookup())
        events = parser.read_events()
        response = ResponseParser.escape_response(response)



        parser.feed("<smah-msg>\n" + response + "\n</smah-msg>")
        for action, elem in events:
            if action == "end":
                if isinstance(elem, ExecTag):
                    ResponseParser.replace_tag(elem, replace=elem.markdown, tail=elem.tail)
                elif isinstance(elem, SetConditionTag):
                    ResponseParser.replace_tag(elem, replace=None, tail=elem.tail)
                elif isinstance(elem, ThoughtTag):
                    ResponseParser.replace_tag(elem, replace=(None if options.get("strip-cot", True) else elem.markdown), tail=elem.tail)

        root = parser.close()
        response = etree.tostring(root, pretty_print=True).decode()
        response = ResponseParser.unescape_response(response)


        return response[len("<smah-msg>\n"):-(len("\n</smah-msg>") + 1)]






import textwrap
from typing import Optional

from rich.prompt import Prompt, Confirm
from smah.console import std_console
class User:
    CONFIG_VSN = "0.0.1"

    @staticmethod
    def config_vsn():
        return User.CONFIG_VSN

    @staticmethod
    def vsn_supported(vsn):
        if vsn is None:
            return False
        else:
            return vsn <= User.config_vsn()

    def __init__(self, config_data = None):
        config_data = config_data or {}
        self.name: Optional[str] = config_data.get("name")
        self.system_admin_experience: Optional[str] = config_data.get("system_admin_experience")
        self.role: Optional[str] = config_data.get("role")
        self.about: Optional[str] = config_data.get("about")
        self.vsn: Optional[str] = config_data.get("vsn")

    def is_configured(self) -> bool:
        if not self.name:
            return False
        if not self.system_admin_experience:
            return False
        if not self.role:
            return False
        if not self.about:
            return False
        return True

    def to_yaml(self, options=None) -> dict:
        return {
            "vsn": self.config_vsn(),
            "name": self.name,
            "system_admin_experience": self.system_admin_experience,
            "role": self.role,
            "about": self.about,
        }

    def show(self, options=None):
        about = self.about or ""
        about = textwrap.indent(about, "   ")
        template = textwrap.dedent(
            """
            - name: {name}
            - role: {role}
            - system_admin_experience: {experience}
            - about:
            {about}
            """
        ).strip().format(
            name=self.name,
            role=self.role,
            experience=self.system_admin_experience,
            about=about,
        )
        return template


from .inference import inference_terminal_configurator
from .settings import Settings
from .user import User, user_terminal_configurator
from .system import System, system_terminal_configurator
from smah.console import std_console, err_console
from rich.markdown import Markdown

def configurator(settings: Settings, gui=False) -> Settings:
    """
    Interactive Setup
    """
    if gui:
        return terminal_configurator(settings)
    else:
        return terminal_configurator(settings)

def terminal_configurator(settings: Settings) -> Settings:
    std_console.print(Markdown("# Configure"))

    settings.user = user_terminal_configurator(settings.user)
    settings.system = system_terminal_configurator(settings.system)
    settings.inference = inference_terminal_configurator(settings.inference)

    settings.save()
    return settings


import os
import textwrap

import yaml
import logging
from typing import Optional

from rich.markdown import Markdown
from smah.console import err_console
from smah.settings.user import User
from smah.settings.system import System
from smah.settings.inference import Inference

class Settings:
    CONFIG_VSN = "0.0.1"
    DEFAULT_CONFIG_FILE = os.path.expanduser("~/.smah/config.yaml")

    @staticmethod
    def config_vsn() -> str:
        return Settings.CONFIG_VSN

    @staticmethod
    def default_config() -> str:
        return Settings.DEFAULT_CONFIG_FILE

    @staticmethod
    def vsn_supported(vsn: Optional[str]) -> bool:
        """
        Checks if the provided version is supported.

        Args:
            vsn (Optional[str]): The version string to check.

        Returns:
            bool: True if the version is supported, False otherwise.
        """
        if not vsn:
            return False
        return vsn <= Settings.config_vsn()

    def __init__(self, config = None):
        self.vsn: Optional[str] = None
        self.config: str = config or self.default_config()
        self.user: Optional[User] = None
        self.system: Optional[System] = None
        self.inference: Optional[Inference] = None
        self.load()

    def is_configured(self):
        """
        Checks if the settings are fully configured.

        Returns:
            bool: True if the settings are configured, False otherwise.
        """
        if not self.user or not self.user.is_configured():
            return False
        if not self.system or not self.system.is_configured():
            return False
        if not self.inference or not self.inference.is_configured():
            return False
        return True

    def load(self) -> None:
        """
        Loads the profile from the profile path and sets the configuration values.
        """
        if os.path.exists(self.config):
            try:
                with open(self.config, 'r') as file:
                    config_data = yaml.safe_load(file)
                    vsn = config_data.get("vsn")
                    if self.vsn_supported(vsn):
                        self.vsn = vsn
                        self.user = User(config_data.get("user"))
                        self.system = System(config_data.get("system"))
                        self.inference = Inference(config_data.get("inference"))
                    else:
                        logging.error(f"Config version {vsn} is not supported by this version of SMAH")
                        raise Exception(f"Config version {vsn} is not supported by this version of SMAH")
            except Exception as e:
                logging.error(f"Failed to load config: {str(e)}")
                raise e
        else:
            logging.error(f"Config file not found: {self.config}")

    def to_yaml(self, options = None) -> dict:
        """
        Returns the settings as a YAML dictionary.

        Returns:
            dict: The settings as a YAML dictionary.
        """
        return {
            "vsn": self.config_vsn(),
            "user": self.user.to_yaml(options=options) if self.user else None,
            "system": self.system.to_yaml(options=options) if self.system else None,
            "inference": self.inference.to_yaml(options=options) if self.inference else None
        }

    def save(self) -> None:
        """
        Saves the current settings to the profile path.
        """
        try:
            os.makedirs(os.path.dirname(self.config), exist_ok=True)
            with open(self.config, 'w') as file:
                yaml_content = yaml.dump(self.to_yaml({ "save": True }), sort_keys=False)
                file.write(yaml_content)
        except Exception as e:
            raise RuntimeError(f"Failed to save profile: {str(e)}")

    def log(self,
            level: int = logging.DEBUG,
            format: bool = True,
            print: bool = False
            ) -> None:
        """
        Log settings and optionally print to stdout.

        Args:
            format (bool): Flag to enable/disable formatting of settings when printing.
            print (bool): Flag to enable/disable printing of settings.
        """
        try:
            settings_yaml = yaml.dump(self.to_yaml({"stats": True, "save": True}), sort_keys=False)
            logging.log(level, "Settings YAML: %s", settings_yaml)

            if print:
                o = textwrap.dedent(
                    """
                    
                    
                    Settings
                    ========
                    ```yaml
                    {settings_yaml}
                    ```
                    """
                ).strip().format(settings_yaml=settings_yaml)
                if format:
                    o = Markdown(o)
                    err_console.print(o)
                else:
                    err_console.print(o)
        except Exception as e:
            logging.error("Exception raised while logging settings: %s", str(e))


def up(cursor):
    """
    Apply schema.
    """
    cursor.execute(
        """
        CREATE TABLE IF NOT EXISTS chat_history(
            id INTEGER PRIMARY KEY AUTOINCREMENT,
            title TEXT,
            created_on TIMESTAMP DEFAULT CURRENT_TIMESTAMP,
            modified_on TIMESTAMP DEFAULT CURRENT_TIMESTAMP
        )
        """
    )
    cursor.execute(
        """
        CREATE TABLE IF NOT EXISTS chat_history_details(
            chat_history_id INTEGER PRIMARY KEY,
            args JSON,
            plan JSON,
            pipe_input TEXT DEFAULT NULL,
            FOREIGN KEY(chat_history_id) REFERENCES chat_history(id)
        )
        """
    )
    cursor.execute(
        """
        CREATE TABLE IF NOT EXISTS chat_history_message(
            id INTEGER PRIMARY KEY AUTOINCREMENT,
            chat_history_id INTEGER,
            message JSON,
            created_on TIMESTAMP DEFAULT CURRENT_TIMESTAMP,
            FOREIGN KEY(chat_history_id) REFERENCES chat_history(id)
        )
        """
    )
    cursor.execute(
        """
        CREATE TABLE IF NOT EXISTS settings(
            setting VARCHAR(32) PRIMARY KEY,
            setting_value VARCHAR(255)
        )
        """
    )


def down(cursor):
    """
    Rollback schema.
    """
    cursor.execute("DROP TABLE settings")
    cursor.execute("DROP TABLE chat_history_details")
    cursor.execute("DROP TABLE chat_history_message")
    cursor.execute("DROP TABLE chat_history")


from typing import Optional

import rich.box
from rich.markdown import Markdown
from rich.panel import Panel
from rich.prompt import Confirm, Prompt

from .operating_system import OperatingSystem, operating_system_terminal_configurator
from .system import System

from smah.console import std_console, err_console, prompt_choice, prompt_string
def system_terminal_configurator(system: Optional[System]) -> System:
    std_console.print(Markdown("## Configure System Settings"))
    system = system or System()

    if system.is_configured():
        show(system)
        if not Confirm.ask("edit?", default=False):
            return system

    while True:
        system = prompt(system)

        std_console.print("\n")
        show(system, label = "Confirm System Settings", border_style="green bold")
        if Confirm.ask("confirm?", default=True):
            return system

def prompt(system: System) -> System:
    system.operating_system = operating_system_terminal_configurator(system.operating_system)
    return system

def show(system: System, label = "System Settings", border_style="white bold"):
    body = system.show()
    template = Markdown(body)
    template = Panel(template, title=label, border_style=border_style, box=rich.box.ROUNDED)
    std_console.print(template)


import json
import logging
import subprocess
import textwrap

import rich.box
from typing import Optional, Tuple

import yaml
from openai import OpenAI, NotGiven, NOT_GIVEN
from openai.types.chat import ChatCompletion
from rich.markdown import Markdown
from rich.panel import Panel
from rich.prompt import Prompt, Confirm

from smah.console import std_console, err_console
from smah.runner.response_parser import ResponseParser
from smah.settings.inference.provider.model import Model
from smah.runner.prompts import Prompts
from smah.database import Database

class Runner:
    MAX_PIPE_LENGTH = 2048
    PIPE_HEAD_LENGTH = 1024


    @staticmethod
    def log_query_plan(plan: dict, level: int = logging.DEBUG, show: bool = False):
        plan = yaml.dump(plan, sort_keys=False)
        logging.log(level, f"Query Plan:\n{plan}")
        if show:
            err_console.print(Panel(
                Markdown("```yaml\n" + plan + "\n```\n"),
                title="Query Plan",
                style="bold yellow",
                box=rich.box.SQUARE)
            )

    @staticmethod
    def log_pipe_plan(plan: dict, level: int = logging.DEBUG, show: bool = False):
        plan = yaml.dump(plan, sort_keys=False)
        logging.log(level, f"Pipe Plan:\n{plan}")
        if show:
            err_console.print(Panel(
                Markdown("```yaml\n" + plan + "\n```\n"),
                title="Pipe Plan",
                style="bold yellow",
                box=rich.box.SQUARE)
            )

    @staticmethod
    def log_mode(mode: str, level: int = logging.INFO, show: bool = False) -> None:
        logging.log(level, f"Processing In {mode} Mode")
        if show:
            err_console.print(Panel(
                f"Processing In {mode} Mode",
                title="Mode",
                style="bold white",
                box=rich.box.ROUNDED)
            )


    @staticmethod
    def log_openai_completion_request(
            model: Model,
            thread: list,
            response_format: dict | NotGiven,
            options: Optional[dict] = None,
            show: bool = False,
            level: int = logging.INFO
    ) -> None:
        payload = yaml.dump(
            {
                'model': model.to_yaml(),
                'response_format': response_format if response_format != NOT_GIVEN else False,
                'options': options or False,
                'thread': thread
            },
            sort_keys=False
        )
        logging.log(level, f"OpenAI Completion Payload:\n{payload}")
        if show:
            err_console.print(Panel(
                Markdown("```yaml\n" + payload + "\n```\n"),
                title="OpenAI Completion Payload",
                style="bold white",
                box=rich.box.ROUNDED)
            )

    @staticmethod
    def log_openai_completion_response(response: ChatCompletion, level = logging.INFO, show: bool = False) -> None:
        payload = yaml.dump(
            response,
            sort_keys=False
        )
        logging.log(level, f"OpenAI Completion Response:\n{payload}")
        if show:
            err_console.print(Panel(
                Markdown("```yaml\n" + payload + "\n```\n"),
                title="OpenAI Completion Response",
                style="bold white",
                box=rich.box.ROUNDED)
            )

    @staticmethod
    def planner_response(response: ChatCompletion) -> Tuple[bool, dict] | None:
        plan = json.loads(response.choices[0].message.content)
        required_keys = ["title", "model", "reason", "include_settings", "include_settings_reason", "format_output",
                         "format_output_reason", "instructions"]
        if all(key in plan for key in required_keys):
            return True, plan
        else:
            missing_keys = [key for key in required_keys if key not in plan]
            logging.error(f"Missing keys: {missing_keys}")
            return None

    def __init__(self, args, settings):
        self.args = args
        self.settings = settings
        self.db = Database(args)




    def openai_client(self):
        api_key = self.settings.inference.providers['openai'].api_key(self.args)
        client = OpenAI(
            api_key=api_key
        )
        return client

    @staticmethod
    def replace_exec_tags(content: str):
        # https://lxml.de/element_classes.html
        # find <exec tags and extract the shell attribute and inner title, purpose and command tags using regex
        # replace the exec tag with the output of the command
        # return the content
        pass




    @staticmethod
    def print_message(message: dict, format: bool = False, strip_cot = True,styles: Optional[dict] = None):


        if format:
            styles = styles or {
                'assistant': 'bold white',
                'user': 'bold blue',
                'default': 'bold green'
            }
            style = styles.get(message['role'], styles.get('default','bold green'))
            content = message['content']
            content = ResponseParser.to_markdown(content, {'strip-cot': strip_cot})
            std_console.print(
                Panel(Markdown(content, style="white"), title=message['role'], style=style, box=rich.box.ROUNDED)
            )
        else:
            std_console.print(f"\n\n--- {message['role']} ---")
            std_console.print(message['content'])



    def resume(self, id: int, title: str, plan: dict, pipe: str, messages: list) -> None:
        model_name = self.args.model or plan['model']
        model = self.settings.inference.models[model_name]
        open = textwrap.dedent(
            f"""
            Continue Session #{id} - {title} ({model.provider}.{model.model})
            =========
            """
        )
        std_console.print(Markdown(open) if self.args.rich else open)

        thread = [
            Prompts.conventions(),
            Prompts.ack(),
            Prompts.system_settings(self.settings, include_system=plan['include_settings']),
            Prompts.ack(),
        ]

        if pipe:
            thread.append(Prompts.message(content=f"--- INPUT ---\n{pipe}"))
            thread.append(Prompts.ack())

        for message in messages:
            thread.append(Prompts.message(content=message['content'], role=message['role']))
            self.print_message(message, format=self.args.rich)

        query = Prompt.ask("[bold green]Message[/bold green]: (type 'exit' or enter to end session)")
        query = query.strip()
        while query != 'exit' and query:
            query_message = Prompts.message(content=query, role='user')
            self.print_message(query_message, format=self.args.rich, strip_cot=False)

            # Query with Instructions
            thread.append(Prompts.query_prompt(request=query))
            response = self.run(model, thread)

            # Response
            message = Prompts.message(role=response.choices[0].message.role, content=response.choices[0].message.content)
            self.print_message(message, format=self.args.rich)

            # Extract Commands
            commands = ResponseParser.extract_commands(response.choices[0].message.content) or []
            for command in commands:
                std_console.print(
                    Panel(
                        Markdown(
                            textwrap.dedent(
                                """
                                `RUNNING SHELL COMMANDS MAY BE DANGEROUS: BE CAREFUL`
                                
                                title: 
                                {title}
                                
                                purpose: 
                                {purpose}
    
                                ```{shell} 
                                {command} 
                                ```                       
                                """
                            ).format(
                                title=command['title'],
                                purpose=command['purpose'],
                                command=command['command'],
                                shell=command['shell']
                            ),
                            style="white"
                        ),
                        title="EXEC COMMAND",
                        style="bold red",
                        box=rich.box.ROUNDED
                    )
                )
                c = Confirm.ask("[bold green]execute?[/bold green]")
                if c:
                    # This is dangerous
                    subprocess.run(command['command'], shell=True)

            # Update Chat History
            self.db.append_to_chat(id, [query_message, message])

            # Continue
            query = Prompt.ask("[bold green]Message[/bold green]: (type 'exit' or enter to end session)")

        exit(0)

    def run(self,
            model: Model,
            thread: list,
            response_format: dict | NotGiven = NOT_GIVEN,
            tools: dict | NotGiven = NOT_GIVEN,
            options: Optional[dict] = None,
            show: bool = False
            ):
        options = options or {}
        if model.provider == "openai":
            self.log_openai_completion_request(
                model=model,
                thread=thread,
                response_format=response_format,
                options=options,
                show=show
                )

            client = self.openai_client()
            model_settings = model.settings or {}

            max_output_tokens = model.context.get("out", 4096)
            max_tokens = options.get("max_tokens", model_settings.get("max_tokens", max_output_tokens))
            max_completion_tokens = options.get("max_completion_tokens", model.settings.get("max_completion_tokens", max_tokens))

            if model_settings.get("max_completion_tokens"):
                max_tokens = NOT_GIVEN
            else:
                max_completion_tokens = NOT_GIVEN

            response = client.chat.completions.create(
                model=model.model,
                messages=thread,
                max_completion_tokens=max_completion_tokens,
                max_tokens=max_tokens,
                response_format=response_format,
                tools=tools
            )
            self.log_openai_completion_response(response, show=show)

            return response



    def inference_model(self, task: str) -> Optional[Model]:
        model = None
        if task == 'query':
            model = self.settings.inference.models.get(self.args.model_query or self.args.model)
        if task == 'pipe':
            model = self.settings.inference.models.get(self.args.model_pipe or self.args.model)
        if task == 'interactive':
            model = self.settings.inference.models.get(self.args.model_interactive or self.args.model)
        if task == 'edit':
            model = self.settings.inference.models.get(self.args.model_edit or self.args.model)
        if task == 'review':
            model = self.settings.inference.models.get(self.args.model_review or self.args.model)

        if model:
            return model

        l = self.settings.inference.model_picker.get(task) or self.settings.inference.model_picker.get('default') or []
        for key in l:
            if key in self.settings.inference.models:
                return self.settings.inference.models[key]
        return None

    def query_plan(self, query: str) -> Optional[Tuple[bool, dict]]:
        self.log_mode("Query Plan", show=self.args.verbose >= 1)
        planner = self.inference_model("query")
        response = self.run(
            model=planner,
            thread=[
                Prompts.conventions(),
                Prompts.ack(),
                Prompts.system_settings(self.settings),
                Prompts.ack(),
                Prompts.select_model(self.settings.inference, request=query),
            ],
            response_format=Prompts.planner_response_format()
        )
        return self.planner_response(response)

    def pipe_plan(self, query: str, pipe: str) -> Optional[Tuple[bool, dict]]:
        self.log_mode("Pipe Plan", show=self.args.verbose >= 1)
        planner = self.inference_model("pipe")
        request = Prompts.pipe_request(query, pipe)

        thread = [
            Prompts.conventions(),
            Prompts.ack(),
            Prompts.system_settings(self.settings),
            Prompts.ack(),
            Prompts.select_model(
                self.settings.inference,
                request=request,
                additional_instructions="This is a pipe input processing request. Unless asked for formatted output assume desired output is to be raw terminal output."
            )
        ]
        response = self.run(
            model=planner,
            thread=thread,
            response_format=Prompts.planner_response_format()
        )
        return self.planner_response(response)




    def query(self, query: str) -> Optional[str]:
        self.log_mode("Query", show=self.args.verbose >= 1)
        plan = self.query_plan(query)
        if plan:
            _, p = plan
            self.log_query_plan(p, show=self.args.verbose >= 2)

            request = textwrap.dedent(
                """\
                {request}
                
                Additional Instructions:
                
                {instructions}
                """).format(request=query, instructions=p["instructions"])
            model = self.settings.inference.models[p["model"]]
            print(query)
            self.print_message(Prompts.message(content=request), format=self.args.rich, strip_cot=False)

            response = self.run(
                model=model,
                thread=[
                    Prompts.conventions(),
                    Prompts.ack(),
                    Prompts.system_settings(self.settings, include_system=p["include_settings"]),
                    Prompts.ack(),
                    Prompts.query_prompt(request=request)
                ]
            )


            msg = {'role': 'assistant', 'content': response.choices[0].message.content}
            self.print_message(msg, format=self.args.rich)

            # Extract Commands
            commands = ResponseParser.extract_commands(response.choices[0].message.content) or []
            for command in commands:
                std_console.print(
                    Panel(
                        Markdown(
                            textwrap.dedent(
                                """
                                `RUNNING SHELL COMMANDS MAY BE DANGEROUS: BE CAREFUL`
                                
                                title: 
                                {title}

                                purpose: 
                                {purpose}

                                ```{shell} 
                                {command} 
                                ```                       
                                """
                            ).format(
                                title=command['title'],
                                purpose=command['purpose'],
                                command=command['command'],
                                shell=command['shell']
                            ),
                            style="white"
                        ),
                        title="EXEC COMMAND",
                        style="bold red",
                        box=rich.box.ROUNDED
                    )
                )
                c = Confirm.ask("[bold green]execute?[/bold green]")
                if c:
                    # This is dangerous
                    subprocess.run(command['command'], shell=True)



            self.db.save_chat(
                p["title"],
                self.args,
                p,
                [
                    Prompts.message(content=request),
                    {'role': 'assistant', 'content': response.choices[0].message.content}
                ]
            )



            return response.choices[0].message.content
        return None


    def pipe(self, query: str, pipe: str) -> str | None:
        plan = self.pipe_plan(query,pipe)
        if plan:
            _, p = plan
            self.log_pipe_plan(p, show=self.args.verbose >= 2)

            request = textwrap.dedent(
                """\
                {query}
                
                Additional Instructions:
                
                {instructions}
                --- INPUT ---
                {pipe}
                """
            ).format(
                query=textwrap.dedent(query),
                pipe=pipe,
                instructions=p["instructions"]
            )

            model = self.settings.inference.models[p["model"]]
            response = self.run(
                model=model,
                thread=[
                    Prompts.conventions(),
                    Prompts.ack(),
                    Prompts.system_settings(self.settings, include_system=p["include_settings"]),
                    Prompts.ack(),
                    Prompts.pipe_prompt(),
                    Prompts.ack(),
                    Prompts.message(content=request),
                ]
            )
            response_body = response.choices[0].message.content
            if p["format_output"] and self.args.rich:
                std_console.print(Markdown(response_body))
            else:
                print(response_body)

            request = textwrap.dedent(
                """\
                {request}                

                Additional Instructions:
                
                {instructions}
                """).format(request=query, instructions=p["instructions"])
            self.db.save_chat(
                p["title"],
                self.args,
                p,
                [
                    Prompts.message(content=request),
                    {'role': 'assistant', 'content': response.choices[0].message.content}
                ],
                pipe=pipe
            )
            return response.choices[0].message.content
        return None


    def interactive(self, query: Optional[str] = None, pipe: Optional[str] = None) -> None:
        self.log_mode("Interactive", show=self.args.verbose >= 1)


from .base_stats import BaseStats
import textwrap
import psutil
import datetime

class DiskStats(BaseStats):
    """
    Represents disk statistics.

    Attributes:
        time_stamp (datetime): The timestamp of the last update.
        total (float): The last recorded total disk space.
        available (float): The last recorded available disk space.
        used (float): The last recorded used disk space.
        percent (float): The last recorded disk usage percentage.
    """

    @staticmethod
    def disk_info(reading):
        """
        Retrieves disk information based on the specified reading type.

        Args:
            reading (str): The type of disk information to retrieve.

        Returns:
            float: The requested disk information.
        """

        try:
            if reading == "total":
                return round(psutil.disk_usage('/').total / (1024.0 ** 3), 2)
            elif reading == "free":
                return round(psutil.disk_usage('/').free / (1024.0 ** 3), 2)
            elif reading == "used":
                return round(psutil.disk_usage('/').used / (1024.0 ** 3), 2)
            elif reading == "percent":
                return round(psutil.disk_usage('/').percent, 2)
        except Exception as e:
            return None

    def __init__(self):
        """
        Initializes the DiskStats instance.
        """
        super().__init__()

        self.total = None
        self.free = None
        self.used = None
        self.percent = None

    def update(self):
        """
        Updates the disk statistics.
        """

        self.time_stamp = datetime.datetime.now()
        self.total = self.disk_info("total")
        self.free = self.disk_info("free")
        self.used = self.disk_info("used")
        self.percent = self.disk_info("percent")

    def readings(self, threshold=100):
        """
        Retrieves the current disk readings, updating if necessary.

        Args:
            threshold (int): The threshold in microseconds.

        Returns:
            dict: The current disk readings.
        """

        if self.stale(threshold):
            self.update()
        return {
            "time": self.time_stamp,
            "total": self.total,
            "free": self.free,
            "used": self.used,
            "percent": self.percent
        }

    def show(self, options=None):
        if self.stale():
            self.update()
        template = textwrap.dedent(
            """
            - time: {time}
            - total: {total}
            - free: {free}
            - used: {used}
            - percent: {percent}
            """
        ).strip().format(
            time=self.time_stamp,
            count=self.total,
            free=self.free,
            used=self.used,
            percent=self.percent
        )
        return template